#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test filtering out Python packages already present in the base image."""

from typing import Optional

import pytest

from thoth.adviser.context import Context
from thoth.adviser.enums import RecommendationType
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.sieves import ThothS2IPackagesSieve
from thoth.python import PackageVersion
from thoth.python import Source

from ..base import AdviserUnitTestCase


class TestThothS2IPackagesSieve(AdviserUnitTestCase):
    """Test filtering out packages based on packages present in S2I."""

    UNIT_TESTED = ThothS2IPackagesSieve

    _EXAMPLE_PYTHON_PACKAGES_PRESENT = [
        # S2I packages
        {
            "location": "/opt/app-root/lib/python3.8/site-packages",
            "package_name": "jupyterhub",
            "package_version": "1.4.1",
        },
        {
            "location": "/opt/app-root/lib/python3.8/site-packages",
            "package_name": "wheel",
            "package_version": "0.36.2",
        },
        {
            "location": "/opt/app-root/lib/python3.8/site-packages",
            "package_name": "tensorflow",
            "package_version": "2.5.0",
        },
        {
            "location": "/opt/app-root/lib/python3.8/site-packages",
            "package_name": "numpy",
            "package_version": "1.21.0",
        },
        # System packages
        {
            "location": "/usr/lib64/python3.6/site-packages",
            "package_name": "gpg",
            "package_version": "1.13.1",
        },
        {
            "location": "/usr/lib/python3.6/site-packages",
            "package_name": "requests",
            "package_version": "2.20.0",
        },
    ]

    @pytest.mark.skip(reason="Default configuration is adjusted")
    def test_default_configuration(self) -> None:
        """Skip as default configuration is adjusted."""

    def test_verify_multiple_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        builder_context.recommendation_type = RecommendationType.LATEST

        builder_context.project.runtime_environment.base_image = "quay.io/thoth-station/s2i-thoth-ubi8-py38:v1.0.0"

        builder_context.graph.should_receive("get_last_analysis_document_id").with_args(
            "quay.io/thoth-station/s2i-thoth-ubi8-py38", "1.0.0", is_external=False
        ).and_return("package-extract-foo").once()

        # Return only one package as the test tests for one inclusion.
        builder_context.graph.should_receive("get_python_package_version_all").with_args(
            "package-extract-foo"
        ).and_return([self._EXAMPLE_PYTHON_PACKAGES_PRESENT[0]]).once()

        self.verify_multiple_should_include(builder_context)

    @pytest.mark.parametrize(
        "base_image",
        [
            None,
            "fedora:32",
            "quay.io/thoth-station/s2i-thoth-ubi8-py38",  # No version.
        ],
    )
    def test_no_should_include(self, base_image: Optional[str], builder_context: PipelineBuilderContext) -> None:
        """Test not including this pipeline unit."""
        builder_context.project.runtime_environment.base_image = base_image
        assert list(self.UNIT_TESTED.should_include(builder_context)) == []

    def test_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Test including this pipeline unit."""
        builder_context.project.runtime_environment.base_image = "quay.io/thoth-station/s2i-thoth-ubi8-py38:v1.0.0"

        builder_context.graph.should_receive("get_last_analysis_document_id").with_args(
            "quay.io/thoth-station/s2i-thoth-ubi8-py38", "1.0.0", is_external=False
        ).and_return("package-extract-foo").once()

        builder_context.graph.should_receive("get_python_package_version_all").with_args(
            "package-extract-foo"
        ).and_return(self._EXAMPLE_PYTHON_PACKAGES_PRESENT).once()

        assert list(self.UNIT_TESTED.should_include(builder_context)) == [
            {
                "package_name": "jupyterhub",
                "package_version": "1.4.1",
            },
            {
                "package_name": "wheel",
                "package_version": "0.36.2",
            },
            {
                "package_name": "tensorflow",
                "package_version": "2.5.0",
            },
            {
                "package_name": "numpy",
                "package_version": "1.21.0",
            },
        ]

    def test_sieve(self, context: Context) -> None:
        """Test removing Python packages present in the base image."""
        pypi = Source("https://pypi.org/simple")
        pv1 = PackageVersion(name="tensorflow", version="==2.5.0", index=pypi, develop=False)
        pv2 = PackageVersion(name="tensorflow", version="==2.4.0", index=pypi, develop=False)
        pv3 = PackageVersion(name="tensorflow", version="==2.3.0", index=pypi, develop=False)

        package_versions = [pv1, pv2, pv3]

        unit = self.UNIT_TESTED()
        unit.update_configuration({"package_name": pv2.name, "package_version": pv2.locked_version})
        unit.pre_run()

        with unit.assigned_context(context):
            assert list(unit.run((pv for pv in package_versions))) == [pv2]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test implementation of GitHub release notes wrap."""

from typing import Generator

import flexmock
import pytest
import yaml

from thoth.adviser.context import Context
from thoth.adviser.state import State
from thoth.adviser.prescription.v1 import GHReleaseNotesWrapPrescription
from thoth.adviser.prescription.v1.schema import PRESCRIPTION_GH_RELEASE_NOTES_WRAP_SCHEMA
from thoth.python import PackageVersion
from thoth.python import Source

from .base import AdviserUnitPrescriptionTestCase


class TestGHReleaseNotesWrapPrescription(AdviserUnitPrescriptionTestCase):
    """Tests related to wrap prescription v1."""

    @staticmethod
    def _instantiate_gh_release_notes_wrap(
        prescription_str: str,
    ) -> Generator[GHReleaseNotesWrapPrescription, None, None]:
        """Instantiate GitHub release notes wrap to match desired configuration for testing purposes."""
        prescription = yaml.safe_load(prescription_str)

        PRESCRIPTION_GH_RELEASE_NOTES_WRAP_SCHEMA(prescription)
        GHReleaseNotesWrapPrescription.set_prescription(prescription)

        conf_generator = GHReleaseNotesWrapPrescription.should_include(
            builder_context=flexmock(is_included=lambda _: False)
        )
        for configuration in conf_generator:
            unit = GHReleaseNotesWrapPrescription()
            unit.update_configuration(configuration)
            yield unit

    def test_run_no_resolved(self, context: Context, state: State) -> None:
        """Test running this pipeline unit not matching any resolved dependency."""
        prescription_str = """
name: GHReleaseNotes
type: wrap.GHReleaseNotes
should_include:
  adviser_pipeline: true
match:
  state:
    resolved_dependencies:
      - name: numpy
run:
  release_notes:
    organization: thoth-station
    repository: adviser
    tag_version_prefix: v
"""
        units = list(self._instantiate_gh_release_notes_wrap(prescription_str))
        assert len(units) == 1, "Multiple units created, expected only one based on prescription"
        unit = units[0]

        state.resolved_dependencies.clear()
        state.add_resolved_dependency(("flask", "1.0.0", "https://pypi.org/simple"))
        state.justification.clear()

        unit.pre_run()
        with unit.assigned_context(context):
            assert unit.run(state) is None

        assert state.justification == []

    def test_run_add_justification(self, context: Context, state: State) -> None:
        """Test running this pipeline unit resulting in a justification addition."""
        prescription_str = """
name: GHReleaseNotes
type: wrap.GHReleaseNotes
should_include:
  adviser_pipeline: true
match:
  - state:
     resolved_dependencies:
       - name: thoth-adviser
run:
  release_notes:
    organization: thoth-station
    repository: adviser
    tag_version_prefix: v
"""
        units = list(self._instantiate_gh_release_notes_wrap(prescription_str))
        assert len(units) == 1, "Multiple units created, expected only one based on prescription"
        unit = units[0]

        state.resolved_dependencies.clear()
        state.add_resolved_dependency(("flask", "1.0.0", "https://pypi.org/simple"))
        state.add_resolved_dependency(("thoth-adviser", "1.0.0", "https://pypi.org/simple"))
        state.justification.clear()

        unit.pre_run()
        with unit.assigned_context(context):
            assert unit.run(state) is None

        self.verify_justification_schema(state.justification)
        assert set(tuple(i.items()) for i in state.justification) == {
            (
                ("type", "INFO"),
                ("message", "Release notes for package 'thoth-adviser'"),
                ("link", "https://github.com/thoth-station/adviser/releases/tag/v1.0.0"),
                ("package_name", "thoth-adviser"),
            ),
        }

    def test_run_not_index_url(self, context: Context, state: State) -> None:
        """Test running this pipeline unit resulting in a justification addition when "not" index url is used."""
        prescription_str = """
name: GHReleaseNotes
type: wrap.GHReleaseNotes
should_include:
  adviser_pipeline: true
match:
  state:
    resolved_dependencies:
      - name: thoth-solver
        index_url:
          not: https://thoth-station.ninja/simple
run:
  release_notes:
    organization: thoth-station
    repository: solver
"""
        units = list(self._instantiate_gh_release_notes_wrap(prescription_str))
        assert len(units) == 1, "Multiple units created, expected only one based on prescription"
        unit = units[0]

        state.resolved_dependencies.clear()
        state.justification.clear()

        state.add_resolved_dependency(("thoth-solver", "0.5.0", "https://pypi.org/simple"))

        unit.pre_run()
        with unit.assigned_context(context):
            assert unit.run(state) is None

        self.verify_justification_schema(state.justification)
        assert set(tuple(i.items()) for i in state.justification) == {
            (
                ("type", "INFO"),
                ("message", "Release notes for package 'thoth-solver'"),
                ("link", "https://github.com/thoth-station/solver/releases/tag/0.5.0"),
                ("package_name", "thoth-solver"),
            ),
        }

    @pytest.mark.parametrize("develop", [True, False])
    def test_run_develop(self, context: Context, state: State, develop: bool) -> None:
        """Test running this pipeline unit resulting in a justification addition when "not" index url is used."""
        prescription_str = f"""
name: GHReleaseNotes
type: wrap.GHReleaseNotes
should_include:
  adviser_pipeline: true
match:
  state:
    resolved_dependencies:
      name: thoth-solver
      develop: {'true' if develop else 'false'}
run:
  release_notes:
    organization: thoth-station
    repository: solver
    tag_version_prefix: v
"""
        units = list(self._instantiate_gh_release_notes_wrap(prescription_str))
        assert len(units) == 1, "Multiple units created, expected only one based on prescription"
        unit = units[0]

        state.resolved_dependencies.clear()
        state.justification.clear()

        package_version = PackageVersion(
            name="thoth-solver",
            version="==0.5.0",
            index=Source("https://pypi.org/simple"),
            develop=develop,
        )
        state.add_resolved_dependency(package_version.to_tuple())
        context.register_package_version(package_version)

        unit.pre_run()
        with unit.assigned_context(context):
            assert unit.run(state) is None

        self.verify_justification_schema(state.justification)
        assert set(tuple(i.items()) for i in state.justification) == {
            (
                ("type", "INFO"),
                ("message", "Release notes for package 'thoth-solver'"),
                ("link", "https://github.com/thoth-station/solver/releases/tag/v0.5.0"),
                ("package_name", "thoth-solver"),
            ),
        }

    @pytest.mark.parametrize("develop", [True, False])
    def test_run_develop_no_match(self, context: Context, state: State, develop: bool) -> None:
        """Test running this pipeline unit resulting in a justification addition when "not" index url is used."""
        prescription_str = f"""
name: GHReleaseNotes
type: wrap.GHReleaseNotes
should_include:
  adviser_pipeline: true
match:
  state:
    resolved_dependencies:
      name: thoth-solver
      develop: {'true' if develop else 'false'}
run:
  release_notes:
    organization: thoth-station
    repository: solver
    tag_version_prefix: v
"""
        units = list(self._instantiate_gh_release_notes_wrap(prescription_str))
        assert len(units) == 1, "Multiple units created, expected only one based on prescription"
        unit = units[0]

        state.resolved_dependencies.clear()
        state.justification.clear()

        package_version = PackageVersion(
            name="thoth-solver",
            version="==0.5.0",
            index=Source("https://pypi.org/simple"),
            develop=not develop,
        )
        state.add_resolved_dependency(package_version.to_tuple())
        context.register_package_version(package_version)

        unit.pre_run()
        with unit.assigned_context(context):
            assert unit.run(state) is None

        assert not state.justification

    @pytest.mark.parametrize(
        "version_expected,version_provided,include",
        [
            ("~=2.26.0", "2.26.1", True),
            ("~=1.0.0", "2.26.1", False),
        ],
    )
    def test_run_version(
        self, context: Context, state: State, version_expected: str, version_provided: str, include: bool
    ) -> None:
        """Test running this pipeline unit resulting in a justification addition when "not" index url is used."""
        prescription_str = f"""
name: GHReleaseNotes
type: wrap.GHReleaseNotes
should_include:
  adviser_pipeline: true
match:
  state:
    resolved_dependencies:
      name: requests
      version: '{version_expected}'
run:
  release_notes:
    organization: psf
    repository: requests
"""
        units = list(self._instantiate_gh_release_notes_wrap(prescription_str))
        assert len(units) == 1, "Multiple units created, expected only one based on prescription"
        unit = units[0]

        state.resolved_dependencies.clear()
        state.justification.clear()

        package_version = PackageVersion(
            name="requests",
            version=f"=={version_provided}",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )
        state.add_resolved_dependency(package_version.to_tuple())
        context.register_package_version(package_version)

        unit.pre_run()
        with unit.assigned_context(context):
            assert unit.run(state) is None

        assert len(state.justification) == int(include)

    def test_instantiate_multiple(self) -> None:
        """Test instantiating multiple GitHub release notes wraps from a single prescription."""
        prescription_str = f"""
name: GHReleaseNotes
type: wrap.GHReleaseNotes
should_include:
  adviser_pipeline: true
match:
  - state:
      resolved_dependencies:
        name: requests-fork
        index_url: https://pypi.org/simple
  - state:
      resolved_dependencies:
        name: requests
run:
  release_notes:
    organization: psf
    repository: requests
"""
        units = list(self._instantiate_gh_release_notes_wrap(prescription_str))
        assert len(units) == 2
        assert units[0].configuration == {
            "package_name": "requests-fork",
            "package_version": {"index_url": "https://pypi.org/simple", "name": "requests-fork"},
            "release_notes": {"organization": "psf", "repository": "requests"},
        }
        assert units[1].configuration == {
            "package_name": "requests",
            "package_version": {"name": "requests"},
            "release_notes": {"organization": "psf", "repository": "requests"},
        }



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test implementation of pseudonym prescription v1."""

from typing import Optional

import flexmock
import pytest
import yaml

from thoth.adviser.context import Context
from thoth.adviser.prescription.v1 import PseudonymPrescription
from thoth.adviser.prescription.v1.schema import PRESCRIPTION_PSEUDONYM_SCHEMA
from thoth.python import PackageVersion
from thoth.python import Source

from .base import AdviserUnitPrescriptionTestCase


class TestPseudonymPrescription(AdviserUnitPrescriptionTestCase):
    """Tests related to pseudonym prescription v1."""

    def test_run_stack_info(self, context: Context) -> None:
        """Check assigning stack info."""
        prescription_str = """
name: PseudonymUnit
type: pseudonym
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: flask
    version: '>1.0,<=1.1.0'
    index_url: 'https://pypi.org/simple'
run:
  yield:
    package_version:
      name: flask
      locked_version: ==1.2.0
      index_url: 'https://pypi.org/simple'

  stack_info:
    - type: WARNING
      message: Some stack warning message
      link: https://thoth-station.ninja
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_PSEUDONYM_SCHEMA(prescription)
        PseudonymPrescription.set_prescription(prescription)
        package_version = PackageVersion(
            name="flask",
            version="==1.1.0",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )

        context.graph.should_receive("get_solved_python_package_versions_all").with_args(
            package_name="flask",
            package_version="1.2.0",
            index_url="https://pypi.org/simple",
            count=None,
            os_name=context.project.runtime_environment.operating_system.name,
            os_version=context.project.runtime_environment.operating_system.version,
            python_version=context.project.runtime_environment.python_version,
            distinct=True,
            is_missing=False,
        ).and_return([("flask", "1.2.0", "https://pypi.org/simple")]).once()
        self.check_run_stack_info(context, PseudonymPrescription, package_version=package_version)

    @pytest.mark.parametrize("log_level", ["INFO", "ERROR", "WARNING"])
    def test_run_log(self, caplog, context: Context, log_level: str) -> None:
        """Check logging messages."""
        prescription_str = f"""
name: PseudonymUnit
type: pseudonym
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: tensorflow
    version: '~=2.4.0'
    index_url: 'https://pypi.org/simple'
run:
  yield:
    package_version:
      name: intel-tensorflow
      locked_version: ==2.4.0
      index_url: 'https://pypi.org/simple'

  log:
    message: Yet some message logged
    type: {log_level}
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_PSEUDONYM_SCHEMA(prescription)
        PseudonymPrescription.set_prescription(prescription)
        package_version = PackageVersion(
            name="tensorflow",
            version="==2.4.1dev0",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )

        context.graph.should_receive("get_solved_python_package_versions_all").with_args(
            package_name="intel-tensorflow",
            package_version="2.4.0",
            index_url="https://pypi.org/simple",
            count=None,
            os_name=context.project.runtime_environment.operating_system.name,
            os_version=context.project.runtime_environment.operating_system.version,
            python_version=context.project.runtime_environment.python_version,
            distinct=True,
            is_missing=False,
        ).and_return([("intel-tensorflow", "2.4.0", "https://pypi.org/simple")]).once()

        self.check_run_log(caplog, context, log_level, PseudonymPrescription, package_version=package_version)

    @pytest.mark.parametrize("package_name", [None, "flask"])
    def test_run_match(self, package_name: Optional[str], context: Context) -> None:
        """Test proper initialization based on yielded package_name."""
        prescription_str = """
name: PseudonymUnit
type: pseudonym
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: tensorflow-cpu
    index_url: 'https://pypi.org/simple'
run:
  yield:
    package_version:
      name: intel-tensorflow
      locked_version: null
      index_url: 'https://pypi.org/simple'
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_PSEUDONYM_SCHEMA(prescription)
        PseudonymPrescription.set_prescription(prescription)

        package_version = PackageVersion(
            name="tensorflow-cpu",
            version="==2.4.0",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )

        expected_result = [
            ("intel-tensorflow", "2.4.0", "https://pypi.org/simple"),
            ("intel-tensorflow", "2.4.1", "https://pypi.org/simple"),
        ]

        context.graph.should_receive("get_solved_python_package_versions_all").with_args(
            package_name="intel-tensorflow",
            package_version=None,
            index_url="https://pypi.org/simple",
            count=None,
            os_name=context.project.runtime_environment.operating_system.name,
            os_version=context.project.runtime_environment.operating_system.version,
            python_version=context.project.runtime_environment.python_version,
            distinct=True,
            is_missing=False,
        ).and_return(expected_result).once()

        unit = PseudonymPrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            result = list(unit.run(package_version=package_version))

        assert result == expected_result

    def test_should_include_package_name(self) -> None:
        """Test including this pipeline unit."""
        prescription_str = """
name: PseudonymUnit
type: pseudonym
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: tensorflow-cpu
    index_url: 'https://pypi.org/simple'
run:
  yield:
    package_version:
      name: intel-tensorflow
      locked_version: null
      index_url: 'https://pypi.org/simple'
"""
        flexmock(PseudonymPrescription).should_receive("_should_include_base").replace_with(lambda _: True).once()
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_PSEUDONYM_SCHEMA(prescription)
        PseudonymPrescription.set_prescription(prescription)

        builder_context = flexmock()
        assert list(PseudonymPrescription.should_include(builder_context)) == [
            {
                "package_name": "tensorflow-cpu",
                "match": {
                    "package_version": {
                        "name": "tensorflow-cpu",
                        "index_url": "https://pypi.org/simple",
                    },
                },
                "run": {
                    "yield": {
                        "package_version": {
                            "name": "intel-tensorflow",
                            "locked_version": None,
                            "index_url": "https://pypi.org/simple",
                        }
                    }
                },
            }
        ]

    def test_should_include_multi(self) -> None:
        """Test including this pipeline unit multiple times."""
        prescription_str = """
name: PseudonymUnit
type: pseudonym
should_include:
  times: 1
  adviser_pipeline: true
match:
  - package_version:
      name: tensorflow-cpu
  - package_version:
      name: tensorflow-gpu
  - package_version:
      name: intel-tensorflow
  - package_version:
      name: tensorflow
run:
  yield:
    package_version:
      name: intel-tensorflow
      locked_version: null
      index_url: 'https://pypi.org/simple'
"""
        flexmock(PseudonymPrescription).should_receive("_should_include_base").replace_with(lambda _: True).once()
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_PSEUDONYM_SCHEMA(prescription)
        PseudonymPrescription.set_prescription(prescription)

        builder_context = flexmock()
        assert list(PseudonymPrescription.should_include(builder_context)) == [
            {
                "package_name": "tensorflow-cpu",
                "match": {
                    "package_version": {
                        "name": "tensorflow-cpu",
                    }
                },
                "run": {
                    "yield": {
                        "package_version": {
                            "name": "intel-tensorflow",
                            "locked_version": None,
                            "index_url": "https://pypi.org/simple",
                        },
                    },
                },
            },
            {
                "package_name": "tensorflow-gpu",
                "match": {
                    "package_version": {
                        "name": "tensorflow-gpu",
                    }
                },
                "run": {
                    "yield": {
                        "package_version": {
                            "name": "intel-tensorflow",
                            "locked_version": None,
                            "index_url": "https://pypi.org/simple",
                        },
                    },
                },
            },
            {
                "package_name": "intel-tensorflow",
                "match": {
                    "package_version": {
                        "name": "intel-tensorflow",
                    }
                },
                "run": {
                    "yield": {
                        "package_version": {
                            "name": "intel-tensorflow",
                            "locked_version": None,
                            "index_url": "https://pypi.org/simple",
                        },
                    },
                },
            },
            {
                "package_name": "tensorflow",
                "match": {
                    "package_version": {
                        "name": "tensorflow",
                    }
                },
                "run": {
                    "yield": {
                        "package_version": {
                            "name": "intel-tensorflow",
                            "locked_version": None,
                            "index_url": "https://pypi.org/simple",
                        },
                    },
                },
            },
        ]

    def test_no_should_include(self) -> None:
        """Test including this pipeline unit without package name."""
        prescription_str = """
name: PseudonymUnit
type: pseudonym
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: tensorflow-cpu
    index_url: 'https://pypi.org/simple'
run:
  yield:
    package_version:
      name: intel-tensorflow
      locked_version: null
      index_url: 'https://pypi.org/simple'
"""
        flexmock(PseudonymPrescription).should_receive("_should_include_base").replace_with(lambda _: False).once()
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_PSEUDONYM_SCHEMA(prescription)
        PseudonymPrescription.set_prescription(prescription)

        builder_context = flexmock()
        assert list(PseudonymPrescription.should_include(builder_context)) == []

    def test_run_yield_matched_version(self, context: Context) -> None:
        """Check yielding correct version when yield_matched_version is supplied."""
        prescription_str = """
name: PseudonymUnit
type: pseudonym
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: flask
    version: '>1.0,<=1.1.0'
    index_url: 'https://pypi.org/simple'
run:
  yield:
    yield_matched_version: true
    package_version:
      name: flask
      index_url: 'https://pypi.org/simple'
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_PSEUDONYM_SCHEMA(prescription)
        PseudonymPrescription.set_prescription(prescription)
        package_version = PackageVersion(
            name="flask",
            version="==1.1.0",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )

        context.graph.should_receive("get_solved_python_package_versions_all").with_args(
            package_name=prescription["run"]["yield"].get("package_version", {}).get("name"),
            package_version="1.1.0",
            index_url=prescription["run"]["yield"].get("package_version", {}).get("index_url"),
            count=None,
            os_name=context.project.runtime_environment.operating_system.name,
            os_version=context.project.runtime_environment.operating_system.version,
            python_version=context.project.runtime_environment.python_version,
            distinct=True,
            is_missing=False,
        ).and_return([("flask", "1.1.0", "https://pypi.org/simple")]).once()

        unit = PseudonymPrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            result = list(unit.run(package_version))

        assert result == [("flask", "1.1.0", "https://pypi.org/simple")]

    def test_run_yield_matched_version_not_index_url(self, context: Context) -> None:
        """Check yielding correct version when "not" index_url is supplied."""
        prescription_str = """
name: PseudonymUnit
type: pseudonym
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: flask
    version: '>1.0,<=1.1.0'
    index_url:
      not: 'https://pypi.org/simple'
run:
  yield:
    yield_matched_version: true
    package_version:
      name: flask
      index_url: 'https://pypi.org/simple'
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_PSEUDONYM_SCHEMA(prescription)
        PseudonymPrescription.set_prescription(prescription)
        package_version = PackageVersion(
            name="flask",
            version="==1.1.0",
            index=Source("https://thoth-station.ninja/simple"),
            develop=False,
        )

        context.graph.should_receive("get_solved_python_package_versions_all").with_args(
            package_name=prescription["run"]["yield"].get("package_version", {}).get("name"),
            package_version="1.1.0",
            index_url=prescription["run"]["yield"].get("package_version", {}).get("index_url"),
            count=None,
            os_name=context.project.runtime_environment.operating_system.name,
            os_version=context.project.runtime_environment.operating_system.version,
            python_version=context.project.runtime_environment.python_version,
            distinct=True,
            is_missing=False,
        ).and_return([("flask", "1.1.0", "https://pypi.org/simple")]).once()

        unit = PseudonymPrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            result = list(unit.run(package_version))

        assert result == [("flask", "1.1.0", "https://pypi.org/simple")]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A class for adviser's report - output of an adviser run."""

import os
import json
import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import attr

from .pipeline_config import PipelineConfig
from .product import Product


_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class Report:
    """A report stating output of an adviser run."""

    products = attr.ib(type=List[Product], kw_only=True)
    pipeline = attr.ib(type=PipelineConfig, kw_only=True)
    resolver_iterations = attr.ib(type=int, kw_only=True, default=0)
    accepted_final_states_count = attr.ib(type=int, kw_only=True, default=0)
    discarded_final_states_count = attr.ib(type=int, kw_only=True, default=0)
    _stack_info = attr.ib(type=Optional[List[Dict[str, Any]]], kw_only=True, default=None)

    @property
    def stack_info(self) -> Optional[List[Dict[str, Any]]]:
        """Retrieve stack info of this report."""
        return self._stack_info

    def set_stack_info(self, stack_info: List[Dict[str, Any]]) -> None:
        """Set stack information."""
        self._stack_info = stack_info

    def to_dict(self, *, verbose: bool = False) -> Dict[str, Any]:
        """Convert pipeline report to a dict representation."""
        stack_info: List[Dict[str, Any]] = []
        stack_info_metadata = os.getenv("THOTH_ADVISER_METADATA")
        if stack_info_metadata:
            try:
                stack_info = ((json.loads(stack_info_metadata).get("thoth.adviser") or {}).get("stack_info")) or []
            except Exception:
                _LOGGER.exception("Failed to load adviser metadata")

        return {
            "pipeline": self.pipeline.to_dict() if verbose else None,
            "products": [product.to_dict() for product in self.products],
            "stack_info": stack_info + (self._stack_info or []),
            "resolver_iterations": self.resolver_iterations,
            "accepted_final_states_count": self.accepted_final_states_count,
            "discarded_final_states_count": self.discarded_final_states_count,
        }



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test a boot that notifies about labels used."""

from thoth.common import get_justification_link as jl

from thoth.adviser.boots import LabelsBoot
from thoth.adviser.context import Context
from thoth.adviser.enums import RecommendationType
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from ..base import AdviserUnitTestCase


class TestLabelsBoot(AdviserUnitTestCase):
    """Test a boot that notifies about labels used."""

    UNIT_TESTED = LabelsBoot

    def test_verify_multiple_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        builder_context.recommendation_type = RecommendationType.LATEST
        builder_context.project.runtime_environment.should_receive("is_fully_specified").with_args().and_return(True)
        builder_context.labels = {"foo": "bar"}
        self.verify_multiple_should_include(builder_context)

    def test_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Test registering this unit if labels are used."""
        builder_context.labels = {"foo": "bar"}
        assert list(self.UNIT_TESTED.should_include(builder_context)) == [{}]
        builder_context.add_unit(self.UNIT_TESTED())
        assert list(self.UNIT_TESTED.should_include(builder_context)) == []

    def test_should_include_no_labels(self, builder_context: PipelineBuilderContext) -> None:
        """Test not registering this unit if supplied software environment is not fully specified."""
        assert not builder_context.labels
        assert list(self.UNIT_TESTED.should_include(builder_context)) == []

    def test_run(self, context: Context) -> None:
        """Test if the given software environment is solved."""
        context.labels = {"foo": "bar", "baz": "qux"}

        assert not context.stack_info

        unit = self.UNIT_TESTED()
        with unit.assigned_context(context):
            assert unit.run() is None

        assert context.stack_info == [
            {"link": jl("labels"), "message": "Considering label foo=bar in the resolution process", "type": "INFO"},
            {"link": jl("labels"), "message": "Considering label baz=qux in the resolution process", "type": "INFO"},
        ]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A boot unit representatives used for testing.."""

from typing import Any
from typing import Dict
from typing import Generator
from typing import TYPE_CHECKING

from thoth.adviser.boot import Boot
from voluptuous import Schema
from voluptuous import Required

if TYPE_CHECKING:
    from thoth.adviser.pipeline_builder import PipelineBuilderContext


class Boot1(Boot):
    """A testing boot implementation."""

    CONFIGURATION_DEFAULT = {"some_parameter": -0.2, "package_name": "flask"}
    CONFIGURATION_SCHEMA: Schema = Schema({Required("package_name"): str, Required("some_parameter"): float})

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Check if this pipeline unit should be included in the pipeline configuration."""
        yield from ()
        return None

    def run(self) -> None:
        """Run noop method."""


class Boot2(Boot):
    """A testing boot implementation."""

    CONFIGURATION_DEFAULT = {"package_name": None}

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Check if this pipeline unit should be included in the pipeline configuration."""
        yield from ()
        return None

    def run(self) -> None:
        """Run noop method."""


__all__ = ["Boot1", "Boot2"]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Thoth's adviser for recommending Python stacks."""

from .beam import Beam
from .boot import Boot
from .context import Context
from .dependency_monkey import DependencyMonkey
from .dm_report import DependencyMonkeyReport
from .enums import DecisionType
from .enums import Ecosystem
from .enums import PythonRecommendationOutput
from .enums import RecommendationType
from .pipeline_builder import PipelineBuilder
from .pipeline_config import PipelineConfig
from .predictor import Predictor
from .product import Product
from .report import Report
from .resolver import Resolver
from .sieve import Sieve
from .state import State
from .step import Step
from .stride import Stride
from .unit import Unit
from .wrap import Wrap

__title__ = "thoth-adviser"
__version__ = "0.44.0"
__author__ = "Fridolin Pokorny <fridolin@redhat.com>"


__all__ = [
    "Beam",
    "Boot",
    "Context",
    "DecisionType",
    "DependencyMonkey",
    "DependencyMonkeyReport",
    "Ecosystem",
    "PipelineBuilder",
    "PipelineConfig",
    "Predictor",
    "Product",
    "PythonRecommendationOutput",
    "RecommendationType",
    "Report",
    "Resolver",
    "Sieve",
    "State",
    "Step",
    "Stride",
    "__title__",
    "Unit",
    "__version__",
    "Wrap",
]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Boot units implemented in adviser."""

from .fully_specified_environment import FullySpecifiedEnvironment
from .gpu import GPUBoot
from .labels import LabelsBoot
from .pipfile_hash import PipfileHashBoot
from .platform import PlatformBoot
from .prescription_release import PrescriptionReleaseBoot
from .python_version import PythonVersionBoot
from .rhel_version import RHELVersionBoot
from .solved_software_environment import SolvedSoftwareEnvironmentBoot
from .solvers_configured import SolversConfiguredBoot
from .thoth_s2i import ThothS2IBoot
from .thoth_s2i_info import ThothS2IInfoBoot
from .ubi import UbiBoot

# from ._debug import MemTraceBoot


# Relative ordering of units is relevant, as the order specifies order
# in which the asked to be registered - any dependencies between them
# can be mentioned here.
__all__ = [
    # "MemTraceBoot",
    "LabelsBoot",
    "PipfileHashBoot",  # Should be placed before any changes to the input.
    "GPUBoot",  # Should be placed before any GPU specific pipeline unit.
    "ThothS2IBoot",
    "ThothS2IInfoBoot",
    "UbiBoot",
    "PythonVersionBoot",
    "SolvedSoftwareEnvironmentBoot",
    "SolversConfiguredBoot",  # Should be placed after SolvedSoftwareEnvironmentBoot.
    "RHELVersionBoot",
    "PlatformBoot",
    "PrescriptionReleaseBoot",
    "FullySpecifiedEnvironment",
    "SolvedSoftwareEnvironmentBoot",
]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Implementation of Monte-Carlo Tree Search (MCTS) based predictor with adaptive simulated annealing schedule."""

import attr

from typing import Optional
from typing import Tuple
import logging
import math
import os

from ..context import Context
from ..state import State
from .td import TemporalDifference


_LOGGER = logging.getLogger(__name__)
# Percentage relative to `limit` for the heat-up part.
_MCTS_HEAT_UP = int(os.getenv("THOTH_MCTS_HEAT_UP", 10))


@attr.s(slots=True)
class MCTS(TemporalDifference):
    """Implementation of Monte-Carlo Tree Search (MCTS) based predictor with adaptive simulated annealing schedule."""

    _next_state = attr.ib(type=Optional[State], default=None, init=False)

    def pre_run(self) -> None:
        """Initialize pre-running of this predictor."""
        self._next_state = None
        super().pre_run()

    def _temperature_function(self, t0: float, context: Context) -> float:
        """Temperature function used to compute new temperature."""
        # This function, in comparision to TD/SA, does not need to take into account iteration as it
        # works on accepted states.
        k = context.accepted_final_states_count / context.limit
        temperature = t0 * 0.99 ** k
        _LOGGER.debug(
            "New temperature for (iteration=%d, t0=%g, accepted final states=%d, limit=%d, beam size= %d, k=%f) = %g",
            context.iteration,
            t0,
            context.accepted_final_states_count,
            context.limit,
            context.beam.size,
            k,
            temperature,
        )
        return max(temperature, 0.0)

    def set_reward_signal(self, state: State, _: Tuple[str, str, str], reward: float) -> None:
        """Note down reward signal of the last action performed."""
        if math.isnan(reward):
            # Invalid state reached, continue with another one in the next round.
            self._next_state = None
            return
        elif not math.isinf(reward):
            # The state for which we obtained reward is next state, continue with it in the next round.
            self._next_state = state
            return

        # We have reached a final/terminal state - mark down policy we used and accumulated reward.
        total_reward = state.score
        for package_tuple in state.iter_resolved_dependencies():
            record = self._policy.setdefault(package_tuple, [0.0, 0])
            record[0] += total_reward
            record[1] += 1

        # We have reached a new final - get another next time.
        self._next_state = None

    def run(self) -> Tuple[State, Tuple[str, str, str]]:
        """Run MCTS with adaptive simulated annealing schedule."""
        # In a heat up part, we run TD-like learning on the first candidates, not to get stuck in one state
        # that generates its children.
        if self.context.iteration < self.context.limit // _MCTS_HEAT_UP:
            return super().run()

        # As beam can be limited to width, it can happen that the last stack is pushed away (based on the score)
        # from the beam. To avoid expanding state that is not present in the beam, check that the last added state
        # in the beam is the one we keep as next expanded. If they do not match, the last added is not the next state
        # we wanted to expand - this is based on the MCTS logic.
        if self._next_state is not None and self._next_state is self.context.beam.get_last():
            if self.keep_history:
                self._temperature_history.append(
                    (
                        self._temperature,
                        self._next_state is self.context.beam.max(),
                        None,
                        self.context.accepted_final_states_count,
                    )
                )
            return (
                self._next_state,
                self._next_state.get_random_unresolved_dependency(prefer_recent=True),
            )

        return super().run()



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test removing packages based on constraints supplied."""

import pytest

from thoth.adviser.context import Context
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.sieves import ConstraintsSieve
from thoth.python import Constraints
from thoth.python import PackageVersion
from thoth.python import Source

from ..base import AdviserUnitTestCase


class TestConstraintsSieve(AdviserUnitTestCase):
    """Test removing packages not coming from a specific Python package index.."""

    UNIT_TESTED = ConstraintsSieve

    @pytest.mark.skip(reason="The pipeline unit configuration is specific to constraint configuration supplied")
    def test_default_configuration(self, builder_context: PipelineBuilderContext) -> None:
        """Test the default configuration of unit tested."""

    def test_default_environment_marker(self, builder_context: "PipelineBuilderContext") -> None:
        """Test checking the default environment marker configuration."""
        builder_context.project.runtime_environment.python_version = "3.9"
        assert self.UNIT_TESTED.default_environment(builder_context) == {
            "implementation_name": "cpython",
            "implementation_version": "3.9.0",
            "os_name": "posix",
            "platform_machine": "x86_64",
            "platform_python_implementation": "CPython",
            "platform_system": "Linux",
            "python_version": builder_context.project.runtime_environment.python_version,
            "sys_platform": "linux",
            "python_full_version": "3.9.0",
            "platform_version": "",
            "platform_release": "",
        }

    def test_include(self, builder_context: PipelineBuilderContext) -> None:
        """Test including this pipeline unit."""
        builder_context.project.runtime_environment.python_version = "3.9"
        builder_context.project.constraints = Constraints.from_file(
            str(self.data_dir / "constraints" / "constraints_0.txt")
        )
        assert list(self.UNIT_TESTED.should_include(builder_context)) == [
            {"package_name": "numpy", "specifier": "~=2.0"}
        ]

    def test_not_include(self, builder_context: PipelineBuilderContext) -> None:
        """Test not including this pipeline unit."""
        builder_context.project.runtime_environment.python_version = "1.9"
        builder_context.project.constraints = Constraints.from_file(
            str(self.data_dir / "constraints" / "constraints_0.txt")
        )
        assert list(self.UNIT_TESTED.should_include(builder_context)) == []

    def test_verify_multiple_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        builder_context.project.constraints = Constraints.from_file(
            str(self.data_dir / "constraints" / "constraints_0.txt")
        )
        builder_context.project.runtime_environment.python_version = "3.9"
        self.verify_multiple_should_include(builder_context)

    def test_sieve(self, context: "Context") -> None:
        """Test filtering out packages based on constraints configured."""
        package_version_1 = PackageVersion(
            name="pandas", version="==0.0.1", index=Source("https://pypi.org/simple"), develop=True
        )
        package_version_2 = PackageVersion(
            name="pandas", version="==0.0.2", index=Source("https://pypi.org/simple"), develop=True
        )

        unit = self.UNIT_TESTED()
        unit.update_configuration({"package_name": "pandas", "specifier": ">=1.0"})

        with unit.assigned_context(context):
            unit.pre_run()
            assert list(unit.run((pv for pv in (package_version_1, package_version_2)))) == []

    def test_no_sieve(self, context: "Context") -> None:
        """Test not filtering out packages based on constraints configured."""
        package_version = PackageVersion(
            name="pandas", version="==2.0.0", index=Source("https://pypi.org/simple"), develop=True
        )

        unit = self.UNIT_TESTED()
        unit.update_configuration({"package_name": "pandas", "specifier": ">=1.0,<=2.0"})

        with unit.assigned_context(context):
            unit.pre_run()
            assert list(unit.run((pv for pv in (package_version,)))) == [package_version]

    def test_no_sieve_prereleases(self, context: "Context") -> None:
        """Test not filtering pre-releases.

        As Python's packaging handles pre-releases separately, let's check correct handling.
        """
        package_version = PackageVersion(
            name="pandas", version="==2.0.0rc0", index=Source("https://pypi.org/simple"), develop=True
        )

        unit = self.UNIT_TESTED()
        unit.update_configuration({"package_name": "pandas", "specifier": "<=2.0"})

        with unit.assigned_context(context):
            unit.pre_run()
            assert list(unit.run((pv for pv in (package_version,)))) == [package_version]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A base class for implementing package pseudonyms."""

import abc
from typing import Generator
from typing import Tuple

from thoth.python import PackageVersion
from voluptuous import Schema
from voluptuous import Required
import attr

from .unit import Unit


@attr.s(slots=True)
class Pseudonym(Unit):
    """Pseudonym base class implementation."""

    # Pseudonym is always specific to a package.
    CONFIGURATION_SCHEMA: Schema = Schema({Required("package_name"): str})

    @staticmethod
    def is_pseudonym_unit_type() -> bool:
        """Check if this unit is of type pseudonym."""
        return True

    @abc.abstractmethod
    def run(self, package_version: PackageVersion) -> Generator[Tuple[str, str, str], None, None]:
        """Run main entry-point for pseudonyms to map packages to their counterparts."""



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test report as product by a resolver."""

import json
import os
import pytest
from itertools import chain

from thoth.adviser.context import Context
from thoth.adviser.pipeline_config import PipelineConfig
from thoth.adviser.product import Product
from thoth.adviser.report import Report
from thoth.common import RuntimeEnvironment

from .base import AdviserTestCase


@pytest.fixture
def pipeline_config() -> PipelineConfig:  # noqa: D401
    """A fixture for a pipeline config."""
    return PipelineConfig(boots={}, pseudonyms={}, sieves={}, steps={}, strides={}, wraps={})


class TestReport(AdviserTestCase):
    """Test report product by a resolver."""

    def test_stack_info(self, pipeline_config: PipelineConfig) -> None:
        """Test setting a global stack information."""
        report = Report(products=[], pipeline=pipeline_config)
        report.set_stack_info([{"foo": "bar"}])

        assert report.to_dict(verbose=True) == {
            "pipeline": {
                "boots": [],
                "pseudonyms": [],
                "sieves": [],
                "steps": [],
                "strides": [],
                "wraps": [],
            },
            "products": [],
            "stack_info": [{"foo": "bar"}],
            "resolver_iterations": 0,
            "accepted_final_states_count": 0,
            "discarded_final_states_count": 0,
        }

    def test_to_dict(self, context: Context, pipeline_config: PipelineConfig) -> None:
        """Test conversion to a dict."""
        # Reuse project from the context for this test case.
        project = context.project
        project_dict = {"aresto momentum": "avada kedavra"}
        project.should_receive("to_dict").with_args(keep_thoth_section=True).and_return(
            project_dict
        ).twice()  # In test and in the report.

        product = Product(
            project=project,
            score=0.666,
            justification=[{"gryffindor": "le gladium leviosa"}],
            advised_runtime_environment=RuntimeEnvironment.from_dict({"python_version": "3.6"}),
            context=context,
        )

        report = Report(products=[product], pipeline=pipeline_config)

        assert report.to_dict(verbose=True) == {
            "pipeline": pipeline_config.to_dict(),
            "products": [product.to_dict()],
            "stack_info": [],
            "resolver_iterations": 0,
            "accepted_final_states_count": 0,
            "discarded_final_states_count": 0,
        }

    def test_to_dict_metadata(self, context: Context, pipeline_config: PipelineConfig) -> None:
        """Test conversion to a dict with passed metadata."""
        # Reuse project from the context for this test case.
        project = context.project
        project_dict = {"aresto momentum": "avada kedavra"}
        project.should_receive("to_dict").with_args(keep_thoth_section=True).and_return(project_dict)

        product = Product(
            project=project,
            score=0.666,
            justification=[{"gryffindor": "le gladium leviosa"}],
            advised_runtime_environment=RuntimeEnvironment.from_dict({"python_version": "3.6"}),
            context=context,
        )

        stack_info = [{"type": "WARNING", "message": "Hello, metadata"}]

        report = Report(
            products=[product],
            pipeline=pipeline_config,
            stack_info=stack_info,
        )

        stack_info_metadata = {
            "thoth.adviser": {
                "stack_info": stack_info,
            }
        }

        assert "THOTH_ADVISER_METADATA" not in os.environ
        os.environ["THOTH_ADVISER_METADATA"] = json.dumps(stack_info_metadata)

        try:
            assert report.to_dict(verbose=True) == {
                "pipeline": pipeline_config.to_dict(),
                "products": [product.to_dict()],
                "stack_info": list(chain(stack_info, report.stack_info)),
                "resolver_iterations": 0,
                "accepted_final_states_count": 0,
                "discarded_final_states_count": 0,
            }
        finally:
            os.environ.pop("THOTH_ADVISER_METADATA")

    def test_no_verbose(self, context: Context, pipeline_config: PipelineConfig) -> None:
        """Test obtaining report with and without verbose flag set."""
        # Reuse project from the context for this test case.
        project = context.project
        project_dict = {"aresto momentum": "avada kedavra"}
        project.should_receive("to_dict").with_args(keep_thoth_section=True).and_return(project_dict)

        product = Product(
            project=project,
            score=0.666,
            justification=[{"gryffindor": "le gladium leviosa"}],
            advised_runtime_environment=RuntimeEnvironment.from_dict({"python_version": "3.6"}),
            context=context,
        )

        report = Report(
            products=[product],
            pipeline=pipeline_config,
            stack_info=[{"foo": "bar"}],
        )

        assert report.to_dict(verbose=False) == {
            "pipeline": None,
            "products": [product.to_dict()],
            "stack_info": report.stack_info,
            "resolver_iterations": 0,
            "accepted_final_states_count": 0,
            "discarded_final_states_count": 0,
        }

        assert report.to_dict(verbose=True) == {
            "pipeline": pipeline_config.to_dict(),
            "products": [product.to_dict()],
            "stack_info": report.stack_info,
            "resolver_iterations": 0,
            "accepted_final_states_count": 0,
            "discarded_final_states_count": 0,
        }



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A base class for implementing pipeline units - strides and steps."""

import abc
import logging
import os
import re
from typing import Any
from typing import Dict
from typing import Generator
from typing import Optional
from typing import Tuple
from typing import Union
from contextlib import contextmanager

import attr
from voluptuous import Schema
from voluptuous import Required
from voluptuous import Any as SchemaAny
from thoth.python import PackageVersion

from .context import Context
from .exceptions import ParseBaseImageError
from .exceptions import PipelineUnitConfigurationSchemaError
from .dm_report import DependencyMonkeyReport

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .pipeline_builder import PipelineBuilderContext
    from .report import Report


_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class Unit(metaclass=abc.ABCMeta):
    """A base class for implementing pipeline units - strides and steps."""

    _CONTEXT: Optional[Context] = None
    CONFIGURATION_SCHEMA: Schema = Schema({Required("package_name"): SchemaAny(str, None)})
    CONFIGURATION_DEFAULT: Dict[str, Any] = {"package_name": None}

    unit_run = attr.ib(type=bool, default=False, kw_only=True)
    _configuration = attr.ib(type=Dict[str, Any], kw_only=True)

    _RE_CAMEL2SNAKE = re.compile("(?!^)([A-Z]+)")
    _AICOE_PYTHON_PACKAGE_INDEX_URL = "https://tensorflow.pypi.thoth-station.ninja/index/"
    _VALIDATE_UNIT_CONFIGURATION_SCHEMA = bool(int(os.getenv("THOTH_ADVISER_VALIDATE_UNIT_CONFIGURATION_SCHEMA", 1)))
    _DATA_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), "data")

    @classmethod
    def get_unit_name(cls) -> str:
        """Get name of the unit."""
        return cls.__name__

    @staticmethod
    def is_boot_unit_type() -> bool:
        """Check if the unit is of type boot."""
        return False

    @staticmethod
    def is_pseudonym_unit_type() -> bool:
        """Check if the unit is of type pseudonym."""
        return False

    @staticmethod
    def is_sieve_unit_type() -> bool:
        """Check if the unit is of type sieve."""
        return False

    @staticmethod
    def is_step_unit_type() -> bool:
        """Check if the unit is of type step."""
        return False

    @staticmethod
    def is_stride_unit_type() -> bool:
        """Check if the unit is of type step."""
        return False

    @staticmethod
    def is_wrap_unit_type() -> bool:
        """Check if the unit is of type wrap."""
        return False

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Check if the given pipeline unit should be included in the given pipeline configuration."""
        raise NotImplementedError(
            f"Please implement method to register pipeline unit {cls.get_unit_name()!r} to pipeline configuration"
        )

    @classmethod
    @contextmanager
    def assigned_context(cls, context: Context) -> Generator[None, None, None]:
        """Assign context to all units."""
        try:
            cls._CONTEXT = context
            yield
        finally:
            cls._CONTEXT = None

    def __attrs_post_init__(self) -> None:
        """Initialize post-init attributes."""
        # Initialize unit_run always to False so the pipeline unit JSON report can be reused across
        # multiple pipeline unit runs.
        self.unit_run = False

    @_configuration.default
    def _initialize_default_configuration(self) -> Dict[str, Any]:
        """Initialize default unit configuration based on declared class' default configuration."""
        return dict(self.CONFIGURATION_DEFAULT)

    @property
    def context(self) -> Context:
        """Get context in which the unit runs in."""
        if self._CONTEXT is None:
            raise ValueError("Requesting resolver context outside of resolver run")

        return self._CONTEXT

    @property
    def name(self) -> str:
        """Get name of this pipeline unit."""
        return self.__class__.__name__

    @property
    def configuration(self) -> Dict[str, Any]:
        """Get configuration of instantiated pipeline unit."""
        return self._configuration

    def update_configuration(self, configuration_dict: Dict[str, Any]) -> None:
        """Set configuration for a pipeline unit.

        If setting configuration fails due to schema checks, configuration are kept in an invalid state.
        """
        self.configuration.update(configuration_dict)
        if self._VALIDATE_UNIT_CONFIGURATION_SCHEMA and self.CONFIGURATION_SCHEMA:
            _LOGGER.debug("Validating configuration for pipeline unit %r", self.name)
            try:
                self.CONFIGURATION_SCHEMA(self.configuration)
            except Exception as exc:
                _LOGGER.exception(
                    "Failed to validate schema for pipeline unit %r: %s",
                    self.name,
                    exc,
                )
                raise PipelineUnitConfigurationSchemaError(str(exc))

    def to_dict(self) -> Dict[str, Any]:
        """Turn this pipeline unit into its dictionary representation."""
        return {"name": self.name, "configuration": self.configuration, "unit_run": self.unit_run}

    @classmethod
    def is_aicoe_release(cls, package_version: PackageVersion) -> bool:
        """Check if the given package-version is AICoE release."""
        return bool(package_version.index.url.startswith(cls._AICOE_PYTHON_PACKAGE_INDEX_URL))

    @classmethod
    def get_aicoe_configuration(cls, package_version: PackageVersion) -> Optional[Dict[str, Any]]:
        """Get AICoE specific configuration encoded in the AICoE index URL."""
        if not package_version.index.url.startswith(cls._AICOE_PYTHON_PACKAGE_INDEX_URL):
            return None

        index_url = package_version.index.url[len(cls._AICOE_PYTHON_PACKAGE_INDEX_URL) :]
        conf_parts = index_url.strip("/").split("/")  # the last is always "simple"

        if len(conf_parts) == 3:
            # No OS specific release - e.g. manylinux compliant release.
            if not conf_parts[0].startswith("manylinux"):
                _LOGGER.error("Failed to parse a platform tag, unknown AICoE Index URL: %r", package_version.index.url)
                return None

            return {
                "os_name": None,
                "os_version": None,
                "configuration": conf_parts[1],
                "platform_tag": conf_parts[0],
            }
        elif len(conf_parts) == 5:
            # TODO: We have dropped OS-specific builds, so this can go away in future releases...
            if conf_parts[0] != "os":
                _LOGGER.error(
                    "Failed to parse operating system specific URL of AICoE index: %r", package_version.index.url
                )
                return None

            return {
                "os_name": conf_parts[1],
                "os_version": conf_parts[2],
                "configuration": conf_parts[3],
                "platform_tag": None,
            }

        _LOGGER.warning(
            "Failed to parse AICoE specific package source index configuration: %r",
            package_version.index.url,
        )
        return None

    @staticmethod
    def get_base_image(base_image: str, *, raise_on_error: bool = False) -> Optional[Tuple[str, str]]:
        """Return information about base image used."""
        parts = base_image.split(":", maxsplit=1)
        if len(parts) != 2:
            if raise_on_error:
                raise ParseBaseImageError(
                    f"Cannot determine Thoth s2i version information from {base_image}, "
                    "recommendations specific for ABI used will not be taken into account"
                )

            return None

        thoth_s2i_image_name, thoth_s2i_image_version = parts
        if thoth_s2i_image_version.startswith("v"):
            # Not nice as we always prefix with "v" but do not store it with "v" in the database
            # (based on env var exported and detected in Thoth's s2i).
            thoth_s2i_image_version = thoth_s2i_image_version[1:]

        return thoth_s2i_image_name, thoth_s2i_image_version

    def pre_run(self) -> None:  # noqa: D401
        """Called before running any pipeline unit with context already assigned.

        This method should not raise any exception.
        """
        self.unit_run = False

    def post_run(self) -> None:  # noqa: D401
        """Called after the resolution is finished.

        This method should not raise any exception.
        """

    def post_run_report(self, report):
        # type:('Unit', Union[Report, DependencyMonkeyReport]) -> None
        """Post-run method run after the resolving has finished - this method is called only if resolving with a report.

        This method should not raise any exception.
        """



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020-2021 Kevin Postlethwait
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test filtering out Python Packages based on required and available ABI symbols."""

from typing import Optional

import flexmock
import pytest

from thoth.adviser.context import Context
from thoth.adviser.enums import RecommendationType
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.sieves import ThothS2IAbiCompatibilitySieve
from thoth.python import PackageVersion
from thoth.python import Source
from thoth.storages import GraphDatabase

from ..base import AdviserUnitTestCase

_SYSTEM_SYMBOLS = ["GLIBC_2.0", "GLIBC_2.1", "GLIBC_2.2", "GLIBC_2.3", "GLIBC_2.4", "GLIBC_2.5", "GCC_3.4", "X_2.21"]
_REQUIRED_SYMBOLS_A = ["GLIBC_2.9"]
_REQUIRED_SYMBOLS_B = ["GLIBC_2.4"]


class TestThothS2IAbiCompatibilitySieve(AdviserUnitTestCase):
    """Test filtering out packages based on symbols required."""

    UNIT_TESTED = ThothS2IAbiCompatibilitySieve

    def test_verify_multiple_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        builder_context.recommendation_type = RecommendationType.LATEST
        builder_context.project.runtime_environment.base_image = "quay.io/thoth-station/s2i-thoth-ubi8-py38:v0.23.0"
        self.verify_multiple_should_include(builder_context)

    @pytest.mark.parametrize(
        "base_image",
        [
            None,
            "fedora:32",
        ],
    )
    def test_no_should_include(self, base_image: Optional[str], builder_context: PipelineBuilderContext) -> None:
        """Test not including this pipeline unit."""
        builder_context.project.runtime_environment.base_image = base_image
        assert list(self.UNIT_TESTED.should_include(builder_context)) == []

    def test_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Test including this pipeline unit."""
        builder_context.project.runtime_environment.base_image = "quay.io/thoth-station/s2i-thoth-ubi8-py38:v0.23.0"
        assert list(self.UNIT_TESTED.should_include(builder_context)) == [{}]

    def test_no_thoth_s2i_version(self, context: Context) -> None:
        """Test no Thoth S2I version present."""
        context.project.runtime_environment.base_image = "quay.io/thoth-station/s2i-thoth-ubi8-py38"  # No version.

        GraphDatabase.should_receive("get_thoth_s2i_analyzed_image_symbols_all").times(0)

        unit = self.UNIT_TESTED()
        with self.UNIT_TESTED.assigned_context(context):
            unit.pre_run()

        assert not unit.image_symbols

    def test_abi_compat_symbols_present(self, context: Context) -> None:
        """Test if required symbols are correctly identified."""
        source = Source("https://pypi.org/simple")
        package_version = PackageVersion(name="tensorflow", version="==1.9.0", index=source, develop=False)
        flexmock(GraphDatabase)
        GraphDatabase.should_receive("get_thoth_s2i_analyzed_image_symbols_all").and_return(_SYSTEM_SYMBOLS).once()
        GraphDatabase.should_receive("get_python_package_required_symbols").and_return(_REQUIRED_SYMBOLS_B).once()

        context.project.runtime_environment.operating_system.name = "rhel"
        context.project.runtime_environment.operating_system.version = "8.0"
        context.project.runtime_environment.cuda_version = "4.6"
        context.project.runtime_environment.python_version = "3.6"
        context.project.runtime_environment.base_image = "quay.io/thoth-station/s2i-thoth-ubi8-py38:v0.23.0"

        with self.UNIT_TESTED.assigned_context(context):
            sieve = self.UNIT_TESTED()
            sieve.pre_run()
            assert list(sieve.run((p for p in [package_version]))) == [package_version]

    def test_abi_compat_symbols_not_present(self, context: Context) -> None:
        """Test if required symbols being missing is correctly identified."""
        source = Source("https://pypi.org/simple")
        package_version = PackageVersion(name="tensorflow", version="==1.9.0", index=source, develop=False)
        flexmock(GraphDatabase)
        GraphDatabase.should_receive("get_thoth_s2i_analyzed_image_symbols_all").and_return(_SYSTEM_SYMBOLS).once()
        GraphDatabase.should_receive("get_python_package_required_symbols").and_return(_REQUIRED_SYMBOLS_A).once()

        context.project.runtime_environment.operating_system.name = "rhel"
        context.project.runtime_environment.operating_system.version = "8.0"
        context.project.runtime_environment.cuda_version = "4.6"
        context.project.runtime_environment.python_version = "3.6"
        context.project.runtime_environment.base_image = "quay.io/thoth-station/s2i-thoth-ubi8-py38:v0.23.0"

        with self.UNIT_TESTED.assigned_context(context):
            sieve = self.UNIT_TESTED()
            sieve.pre_run()
            assert list(sieve.run((p for p in [package_version]))) == []

    def test_super_pre_run(self, context: Context) -> None:
        """Make sure the pre-run method of the base is called."""
        context.graph.should_receive("get_thoth_s2i_analyzed_image_symbols_all").with_args(
            thoth_s2i_image_name="quay.io/thoth-station/s2i-thoth-ubi8-py38",
            thoth_s2i_image_version="0.23.0",
            is_external=False,
        ).and_return(set()).once()

        context.project.runtime_environment.base_image = "quay.io/thoth-station/s2i-thoth-ubi8-py38:v0.23.0"

        unit = self.UNIT_TESTED()
        assert unit.unit_run is False

        with unit.assigned_context(context):
            unit.pre_run()

        assert unit.unit_run is False



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A sieve that filters out TensorFlow==2.4.0 build as it requires AVX2 instruction set."""

from typing import Any
from typing import Dict
from typing import Generator
from typing import TYPE_CHECKING
import logging

from thoth.common import get_justification_link as jl
from thoth.python import PackageVersion

from ...enums import RecommendationType
from ...sieve import Sieve


if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext


_LOGGER = logging.getLogger(__name__)


class TensorFlow240AVX2IllegalInstructionSieve(Sieve):
    """A sieve that filters out TensorFlow==2.4.0 build as it requires AVX2 instruction set.

    See:
     * https://github.com/tensorflow/tensorflow/issues/45866
     * https://github.com/tensorflow/tensorflow/issues/45744
    """

    CONFIGURATION_DEFAULT = {"package_name": "tensorflow"}

    _JUSTIFICATION_ADDITION = {
        "type": "WARNING",
        "message": "Upstream TensorFlow=2.4.0 build filtered out as it requires AVX2 instruction "
        "set support which is not available.",
        "link": jl("tf_240_avx2"),
    }

    # A tuple (CPU_FAMILY, CPU_MODEL) of Intel processors supporting AVX2:
    #   https://en.wikipedia.org/wiki/Advanced_Vector_Extensions#CPUs_with_AVX2
    #   https://en.wikichip.org/wiki/intel/cpuid
    AVX2_CPUS = frozenset(
        {
            (0x6, 0x5),  # Cascade Lake
            (0x6, 0x6),  # Broadwell, Cannon Lake
            (0x6, 0xA),  # Ice Lake
            (0x6, 0xC),  # Ice Lake, Tiger Lake
            (0x6, 0xD),  # Ice Lake
            (0x6, 0xE),  # Skylake, Kaby Lake, Coffee Lake, Ice Lake, Comet Lake, Whiskey Lake
            (0x6, 0xF),  # Haswell
        }
    )

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Register this pipeline unit for adviser and stable/testing recommendation types."""
        cpu_tuple = (
            builder_context.project.runtime_environment.hardware.cpu_family,
            builder_context.project.runtime_environment.hardware.cpu_model,
        )

        if (
            builder_context.is_adviser_pipeline()
            and not builder_context.is_included(cls)
            and builder_context.recommendation_type != RecommendationType.LATEST
            and all(i is not None for i in cpu_tuple)
            and cpu_tuple not in cls.AVX2_CPUS
        ):
            yield {}
            return None

        yield from ()
        return None

    def run(self, package_versions: Generator[PackageVersion, None, None]) -> Generator[PackageVersion, None, None]:
        """Recommend not to use TensorFlow==2.4.0 on non-AVX2 enabled CPU processors."""
        for package_version in package_versions:
            if (
                package_version.name == "tensorflow"
                and package_version.locked_version == "2.4.0"
                and package_version.index.url == "https://pypi.org/simple"
            ):
                self.context.stack_info.append(self._JUSTIFICATION_ADDITION)
                continue

            yield package_version



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test wrap that sorts justifications for better UX."""

from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.state import State
from thoth.adviser.wraps import SortJustificationsWrap

from ..base import AdviserUnitTestCase


class TestSortPackageNameWrap(AdviserUnitTestCase):
    """Test wrap that sorts justifications for better UX."""

    UNIT_TESTED = SortJustificationsWrap

    def test_verify_multiple_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        self.verify_multiple_should_include(builder_context)

    def test_run_add_justification(self) -> None:
        """Test sorting justifications.."""
        state = State()
        state.add_justification(
            [
                {
                    "type": "INFO",
                    "message": "Info about 'b'",
                    "link": "https://foo/bar",
                    "package_name": "b",
                },
                {
                    "type": "INFO",
                    "message": "Package 'a' in version '1.0.0'",
                    "link": "https://foo/bar",
                    "package_name": "a",
                },
                {
                    "type": "WARNING",
                    "message": "Some warning message printed not specific to any package",
                    "link": "https://foo/bar",
                },
                {
                    "type": "WARNING",
                    "message": "Warning about 'a'",
                    "link": "https://foo/bar",
                    "package_name": "a",
                },
            ]
        )

        unit = self.UNIT_TESTED()
        unit.run(state)

        assert state.justification == [
            {
                "link": "https://foo/bar",
                "message": "Some warning message printed not specific to any package",
                "type": "WARNING",
            },
            {
                "link": "https://foo/bar",
                "message": "Package 'a' in version '1.0.0'",
                "package_name": "a",
                "type": "INFO",
            },
            {"link": "https://foo/bar", "message": "Warning about 'a'", "package_name": "a", "type": "WARNING"},
            {"link": "https://foo/bar", "message": "Info about 'b'", "package_name": "b", "type": "INFO"},
        ]
        self.verify_justification_schema(state.justification)



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A sieve to filter out packages based on Python package index configured."""

import logging
from typing import Dict
from typing import Any
from typing import Generator
from typing import TYPE_CHECKING

import attr
from thoth.python import PackageVersion
from voluptuous import Required
from voluptuous import Schema

from ..sieve import Sieve

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class FilterIndexSieve(Sieve):
    """A sieve to filter out packages based on Python package index configured.

    The filtering is done based on configuration supplied - per package-version. The sieve is
    configured using the following pipeline configuration entry:
      {
        "package_name": "tensorflow",
        "index_url": [
          "https://pypi.org/simple",
          "https://tensorflow.pypi.thoth-station.ninja/index/manylinux2010/AVX2/simple/"
        ]
      }
    """

    CONFIGURATION_DEFAULT = {"package_name": None, "index_url": None}
    CONFIGURATION_SCHEMA = Schema(
        {
            Required("package_name"): str,
            Required("index_url"): [str],
        }
    )

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Register self, never."""
        yield from ()
        return None

    def run(self, package_versions: Generator[PackageVersion, None, None]) -> Generator[PackageVersion, None, None]:
        """Filter out packages based on Python package index configured for the given package."""
        for package_version in package_versions:
            if package_version.name != self.configuration["package_name"]:
                yield package_version

            if package_version.index.url not in self.configuration["index_url"]:
                _LOGGER.warning(
                    "Removing package %r in version %r as index used %r does not conform to any supplied %r",
                    package_version.name,
                    package_version.version,
                    package_version.index.url,
                    self.configuration["index_url"],
                )
                continue

            yield package_version



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Perform a drop out of a new state expansion, randomly."""

import random

from typing import Any
from typing import Generator
from typing import Optional
from typing import Tuple
from typing import List
from typing import Dict
from typing import TYPE_CHECKING
import logging

from thoth.python import PackageVersion
from voluptuous import Required
from voluptuous import Schema

from ..exceptions import NotAcceptable
from ..state import State
from ..step import Step


if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext


_LOGGER = logging.getLogger(__name__)


class DropoutStep(Step):
    """A step that drops a state transition with a certain probability."""

    CONFIGURATION_DEFAULT = {"package_name": None, "probability": 0.9, "multi_package_resolution": False}
    CONFIGURATION_SCHEMA: Schema = Schema(
        {
            Required("package_name"): None,
            Required("cve_penalization"): float,
            Required("multi_package_resolution"): False,
        }
    )

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Do not register the dropout step."""
        yield from ()
        return None

    def run(
        self, state: State, package_version: PackageVersion
    ) -> Optional[Tuple[Optional[float], Optional[List[Dict[str, str]]]]]:
        """Do not accept new state, randomly."""
        if random.random() >= self.configuration["probability"]:
            raise NotAcceptable

        return None



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A sieve for filtering out Python packages that are not part of a new release.

This pipeline unit is registered only if adviser was requested to run to check a new
package release. The pipeline unit filters out all the packages except for the one that
should be included in the stack (the newly released) to check whether there is a stack
with the released package that is better than the one used with an older release.
"""

import os
import logging
from typing import Any
from typing import Dict
from typing import Generator
from typing import Optional
from typing import Set
from typing import Tuple
from typing import TYPE_CHECKING

import attr
import json
from thoth.common import get_justification_link as jl
from thoth.adviser.enums import RecommendationType
from thoth.python import PackageVersion
from voluptuous import Required
from voluptuous import Schema

from ..sieve import Sieve

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class PackageUpdateSieve(Sieve):
    """A sieve for filtering out Python packages that are not part of a new release."""

    CONFIGURATION_DEFAULT = {"package_name": None, "package_version": None, "index_url": None}
    CONFIGURATION_SCHEMA: Schema = Schema(
        {Required("package_name"): str, Required("package_version"): str, Required("index_url"): str}
    )
    _JUSTIFICATION_LINK = jl("update")
    _PACKAGE_UPDATE = os.getenv("THOTH_ADVISER_DEPLOYMENT_PACKAGE_UPDATE")

    _messages_logged = attr.ib(type=Set[Tuple[str, str, str]], factory=set, init=False)
    _package_update = attr.ib(type=Optional[Tuple[str, str, str]], init=False, default=None)

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Include pipeline sieve."""
        if cls._PACKAGE_UPDATE and not builder_context.is_included(cls):
            package_update_dict = json.loads(cls._PACKAGE_UPDATE)

            if builder_context.recommendation_type != RecommendationType.LATEST:
                _LOGGER.warning(
                    "Registering %s sieve pipeline unit when not recommending latest, this error is not critical ...",
                    cls.__name__,
                )

            yield package_update_dict
            return None

        yield from ()
        return None

    def pre_run(self) -> None:
        """Initialize this pipeline unit before each run."""
        self._messages_logged.clear()
        self._package_update = (
            self.configuration["package_name"],
            self.configuration["package_version"],
            self.configuration["index_url"],
        )
        super().pre_run()

    def run(self, package_versions: Generator[PackageVersion, None, None]) -> Generator[PackageVersion, None, None]:
        """Filter out packages that are old releases."""
        for package_version in package_versions:
            package_tuple = package_version.to_tuple()

            if package_tuple != self._package_update:
                if package_tuple not in self._messages_logged:
                    self._messages_logged.add(package_tuple)
                    message = f"Removing package {package_tuple} as advising on a new release"
                    _LOGGER.warning("%s - see %s", message, self._JUSTIFICATION_LINK)
                    self.context.stack_info.append(
                        {
                            "type": "WARNING",
                            "message": message,
                            "link": self._JUSTIFICATION_LINK,
                            "package_name": package_version.name,
                        }
                    )

                continue

            yield package_version



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2018 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# type: ignore

"""Mocking responses from the graph database."""

import os
import typing
from collections import deque

import yaml
from .base import AdviserTestCase


class MockedGraphDatabase:
    """A mocked graph database using YAML files as data."""

    DEFAULT_COUNT = 100

    def __init__(self, database_file: str):
        """Initialize mock database."""
        with open(os.path.join(AdviserTestCase.data_dir, "graph", database_file), "r") as db_file:
            self.db = yaml.safe_load(db_file)
        # A map mapping package id to its tuples - this map can be used after the retrieve
        # transitive_dependencies_python call which fills this dict.
        self._id_map = {}
        self._reverse_id_map = {}

    def connect(self):
        """Connect to fake database (does nothing)."""
        pass

    def get_solved_python_package_versions_all(
        self,
        package_name: str,
        os_name: str = None,
        os_version: str = None,
        python_version: str = None,
        without_error: bool = True,
        count: bool = DEFAULT_COUNT,
        start_offset: bool = 0,
        distinct: bool = True,
        is_missing: bool = False,
    ) -> typing.List[tuple]:
        """Get all versions for the given Python package."""
        result = []
        for version, info in self.db.get(package_name, {}).items():
            for entry in info:
                result.append((package_name, version, entry["index_url"]))

        return result

    def retrieve_transitive_dependencies_python(self, package_name: str, version: str, index: str) -> typing.List[list]:
        """Retrieve transitive dependencies for the given Python package."""
        # Cache for checking whether we have triplet (package_name, version, index) already seen
        # to avoid recursion.
        result = []
        queue = deque()
        queue.append((package_name, version, index))

        while queue:
            package_name, package_version, index = queue.pop()

            if package_name not in self.db:
                raise ValueError(f"The given package {package_name!r} is not present in the database")

            info_entry = self.db[package_name].get(package_version)
            if not info_entry:
                raise ValueError(
                    f"The given package {package_name!r} has no record in the database for version {version!r}"
                )

            for version_entry in info_entry:
                for depends_on_entry in version_entry.get("depends_on", []):
                    for version in depends_on_entry["resolved"]:
                        queue.append(
                            [
                                depends_on_entry["package_name"],
                                version,
                                depends_on_entry["index_url"],
                            ]
                        )

                        source = (package_name, package_version, index)
                        if source in self._reverse_id_map:
                            source_id = self._reverse_id_map[source]
                        else:
                            source_id = id(source)
                            self._id_map[source_id] = source
                            self._reverse_id_map[source] = source_id

                        destination = (
                            depends_on_entry["package_name"],
                            version,
                            depends_on_entry["index_url"],
                        )
                        if destination in self._reverse_id_map:
                            destination_id = self._reverse_id_map[destination]
                        else:
                            destination_id = id(destination)
                            self._id_map[destination_id] = destination
                            self._reverse_id_map[destination] = destination_id

                        result.append((source_id, False, destination_id))

        return result

    def get_python_package_tuples(self, python_package_node_ids: typing.Set[int]) -> typing.Dict[int, tuple]:
        """Get package tuples from fake database."""
        result = {}
        for python_package_id in python_package_node_ids:
            package_tuple = self._id_map.get(python_package_id)
            if package_tuple is None:
                raise ValueError(f"The given id {python_package_id} was never retrieved from the graph")
            result[python_package_id] = package_tuple

        return result



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A boot that recommends to use Thoth's s2i if users do not use it."""

from typing import TYPE_CHECKING
from typing import Generator
from typing import Dict
from typing import Any

from thoth.common import get_justification_link as jl

from ..boot import Boot

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext


class ThothS2IBoot(Boot):
    """A boot that notifies about missing observations."""

    _THOTH_S2I_PREFIX = "quay.io/thoth-station/s2i-thoth-"
    _JUSTIFICATION = [
        {
            "type": "INFO",
            "message": "It is recommended to use Thoth's s2i to have recommendations specific to runtime environment",
            "link": jl("thoth_s2i"),
        }
    ]

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[Any, Any], None, None]:
        """Include this boot in adviser if Thoth s2i is not used.."""
        base_image = builder_context.project.runtime_environment.base_image
        if not builder_context.is_included(cls) and (
            base_image is None or (base_image and not base_image.startswith(cls._THOTH_S2I_PREFIX))
        ):
            yield {}
            return None

        yield from ()
        return None

    def run(self) -> None:
        """Check for no observations made on the given state."""
        self.context.stack_info.extend(self._JUSTIFICATION)



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test a sieve that filters out TensorFlow==2.4.0 build as it requires AVX2 instruction set."""

from typing import Optional

import pytest

from thoth.adviser.context import Context
from thoth.adviser.enums import DecisionType
from thoth.adviser.enums import RecommendationType
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.sieves import TensorFlow240AVX2IllegalInstructionSieve
from thoth.python import PackageVersion
from thoth.python import Source

from ...base import AdviserUnitTestCase


class TestTensorFlowAVX2Step(AdviserUnitTestCase):
    """Test a sieve that filters out TensorFlow==2.4.0 build as it requires AVX2 instruction set."""

    UNIT_TESTED = TensorFlow240AVX2IllegalInstructionSieve

    def test_verify_multiple_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        builder_context.recommendation_type = RecommendationType.STABLE
        # Not None and not in AVX2 CPUs listing.
        builder_context.project.runtime_environment.hardware.cpu_family = 0x0
        builder_context.project.runtime_environment.hardware.cpu_model = 0x0
        self.verify_multiple_should_include(builder_context)

    @pytest.mark.parametrize("recommendation_type", [RecommendationType.STABLE, RecommendationType.TESTING])
    def test_include(self, builder_context: PipelineBuilderContext, recommendation_type: RecommendationType) -> None:
        """Test including this pipeline unit."""
        builder_context.decision_type = None
        builder_context.recommendation_type = recommendation_type
        # Not None and not in AVX2 CPUs listing.
        builder_context.project.runtime_environment.hardware.cpu_family = 0x0
        builder_context.project.runtime_environment.hardware.cpu_model = 0x0
        assert builder_context.is_adviser_pipeline()
        assert list(self.UNIT_TESTED.should_include(builder_context)) == [{}]

    @pytest.mark.parametrize(
        "recommendation_type,decision_type,cpu_family,cpu_model",
        [
            (RecommendationType.STABLE, None, None, None),  # No CPU info.
            (RecommendationType.TESTING, None, 0x6, 0xF),  # AVX2 support in CPU.
            (None, DecisionType.RANDOM, 0x0, 0x0),  # A Dependency Monkey run.
        ],
    )
    def test_no_include(
        self,
        builder_context: PipelineBuilderContext,
        recommendation_type,
        decision_type: DecisionType,
        cpu_family: Optional[int],
        cpu_model: Optional[int],
    ) -> None:
        """Test not including this pipeline unit step."""
        builder_context.decision_type = decision_type
        builder_context.recommendation_type = recommendation_type
        builder_context.project.runtime_environment.hardware.cpu_family = cpu_family
        builder_context.project.runtime_environment.hardware.cpu_model = cpu_model
        assert list(self.UNIT_TESTED.should_include(builder_context)) == []

    def test_tf_avx2(self, context: Context) -> None:
        """Test recommending TensorFlow with AVX2 support."""
        package_version = PackageVersion(
            name="tensorflow",
            version="==2.4.0",
            develop=False,
            index=Source("https://pypi.org/simple"),
        )

        assert not context.stack_info

        with self.UNIT_TESTED.assigned_context(context):
            unit = self.UNIT_TESTED()
            result = list(unit.run((pv for pv in [package_version])))
            assert len(result) == 0

        assert len(context.stack_info) == 1

        justification = context.stack_info[0]
        assert set(justification.keys()) == {"type", "message", "link"}
        assert justification["type"] == "WARNING"
        assert justification["link"]
        assert justification["message"]

    def test_no_tf_avx2(self, context: Context) -> None:
        """Test not recommending TensorFlow without AVX2 support."""
        package_version = PackageVersion(
            name="tensorflow",
            version="==2.4.0",
            develop=False,
            index=Source("https://thoth-station.ninja/simple"),
        )

        assert not context.stack_info

        with self.UNIT_TESTED.assigned_context(context):
            unit = self.UNIT_TESTED()
            result = list(unit.run((pv for pv in [package_version])))
            assert len(result) == 1
            assert result[0] is package_version

        assert not context.stack_info



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A sieve to filter out pre-releases, selectively."""

import logging
from typing import Any
from typing import Dict
from typing import Generator
from typing import Set
from typing import Tuple
from typing import TYPE_CHECKING

import attr
from thoth.common import get_justification_link as jl
from thoth.python import PackageVersion
from voluptuous import Schema
from voluptuous import Required

from ..sieve import Sieve

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class SelectiveCutPreReleasesSieve(Sieve):
    """Enable or disable specific pre-releases for the given set of packages.."""

    CONFIGURATION_DEFAULT = {"package_name": None, "allow_prereleases": None}
    CONFIGURATION_SCHEMA: Schema = Schema(
        {Required("package_name"): None, Required("allow_prereleases"): Schema({str: bool})}
    )
    _JUSTIFICATION_LINK = jl("selective_prereleases")

    packages_seen = attr.ib(type=Set[Tuple[str, str, str]], default=attr.Factory(set), init=False)

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Enable or disable specific pre-releases for the given set of packages.."""
        if builder_context.project.prereleases_allowed:
            if builder_context.project.prereleases_allowed:
                msg = "Ignoring selective pre-releases in [thoth] section as global allow_prereleases flag is set"
                _LOGGER.warning(msg)

            yield from ()
            return None

        if (
            builder_context.is_included(cls)
            or not builder_context.project.pipfile.thoth
            or not builder_context.project.pipfile.thoth.allow_prereleases
        ):
            yield from ()
            return None

        yield {
            "package_name": None,
            "allow_prereleases": builder_context.project.pipfile.thoth.allow_prereleases,
        }

    def pre_run(self) -> None:
        """Initialize this pipeline unit before each run."""
        self.packages_seen.clear()
        super().pre_run()

    def run(self, package_versions: Generator[PackageVersion, None, None]) -> Generator[PackageVersion, None, None]:
        """Cut-off pre-releases if project does not explicitly allows them."""
        for package_version in package_versions:
            if (
                not self.configuration["allow_prereleases"].get(package_version.name, False)
                and package_version.semantic_version.is_prerelease
            ):
                package_tuple = package_version.to_tuple()

                if package_tuple not in self.packages_seen:
                    self.packages_seen.add(package_tuple)
                    msg = f"Removing package {package_tuple} as pre-releases are disabled"
                    _LOGGER.warning("%s - see %s", msg, self._JUSTIFICATION_LINK)
                    self.context.stack_info.append(
                        {
                            "message": msg,
                            "type": "WARNING",
                            "link": self._JUSTIFICATION_LINK,
                        }
                    )

                continue

            yield package_version



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A prescription for boot units."""

import attr

from typing import Any
from typing import Dict
from typing import Generator
from typing import TYPE_CHECKING
from voluptuous import Schema
from voluptuous import Any as SchemaAny
from voluptuous import Required

from .unit import UnitPrescription
from .schema import PRESCRIPTION_BOOT_RUN_SCHEMA
from .schema import PRESCRIPTION_BOOT_MATCH_ENTRY_SCHEMA


if TYPE_CHECKING:
    from ...pipeline_builder import PipelineBuilderContext


@attr.s(slots=True)
class BootPrescription(UnitPrescription):
    """Boot prescription implementation."""

    CONFIGURATION_SCHEMA: Schema = Schema(
        {
            Required("package_name"): SchemaAny(str, None),
            Required("match"): SchemaAny(PRESCRIPTION_BOOT_MATCH_ENTRY_SCHEMA, None),
            Required("run"): PRESCRIPTION_BOOT_RUN_SCHEMA,
        }
    )

    @staticmethod
    def is_boot_unit_type() -> bool:
        """Check if this unit is of type boot."""
        return True

    @staticmethod
    def _yield_should_include(unit_prescription: Dict[str, Any]) -> Generator[Dict[str, Any], None, None]:
        """Yield for every entry stated in the match field."""
        match = unit_prescription.get("match", {})
        run = unit_prescription.get("run", {})
        if isinstance(match, list):
            for item in match:
                yield {"package_name": item.get("package_name"), "match": item, "run": run}
        else:
            yield {"package_name": match.get("package_name") if match else None, "match": match, "run": run}

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Check if the given pipeline unit should be included in the given pipeline configuration."""
        if cls._should_include_base(builder_context):
            prescription: Dict[str, Any] = cls._PRESCRIPTION  # type: ignore
            yield from cls._yield_should_include(prescription)
            return None

        yield from ()
        return None

    def run(self) -> None:
        """Run main entry-point for boot units."""
        super()._run_base()



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test version constraint sieve."""

import pytest

from thoth.adviser.context import Context
from thoth.adviser.sieves import VersionConstraintSieve
from thoth.adviser.exceptions import PipelineUnitConfigurationSchemaError
from thoth.python import Source
from thoth.python import PackageVersion

from ..base import AdviserUnitTestCase


class TestVersionConstrainSieve(AdviserUnitTestCase):
    """Test version constraint sieve."""

    UNIT_TESTED = VersionConstraintSieve

    @pytest.mark.skip(reason="Version constrain sieve is never registered.")
    def test_verify_multiple_should_include(self) -> None:
        """Verify multiple should_include calls do not loop endlessly."""

    def test_configuration_error(self) -> None:
        """Test removing a locked package based on direct dependencies."""
        # Default values should error, no adjustment to values.
        unit = VersionConstraintSieve()
        with pytest.raises(PipelineUnitConfigurationSchemaError):
            unit.update_configuration({})

        unit = VersionConstraintSieve()
        with pytest.raises(PipelineUnitConfigurationSchemaError):
            unit.update_configuration({"package_name": None, "version_specifier": None})

        unit = VersionConstraintSieve()
        with pytest.raises(PipelineUnitConfigurationSchemaError):
            unit.update_configuration(
                {
                    "package_name": "tensorflow",
                    "version_specifier": None,
                }
            )

        unit = VersionConstraintSieve()
        with pytest.raises(PipelineUnitConfigurationSchemaError):
            unit.update_configuration(
                {
                    "package_name": None,
                    "version_specifier": ">2.0",
                }
            )

    def test_default_configuration(self) -> None:
        """Test obtaining default configuration."""
        unit = VersionConstraintSieve()
        assert unit.configuration == {"package_name": None, "version_specifier": None}

    def test_run_filter(self) -> None:
        """Test filtering a package based on version specifier."""
        package_version = PackageVersion(
            name="tensorflow",
            version="==2.0.0",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )
        unit = VersionConstraintSieve()
        unit.update_configuration(
            {
                "package_name": "tensorflow",
                "version_specifier": "<2.0",
            }
        )
        unit.pre_run()
        assert list(unit.run([package_version])) == []

    def test_run_no_filter(self) -> None:
        """Test not filtering a package based on version specifier."""
        package_version = PackageVersion(
            name="tensorboard",
            version="==2.1.0",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )
        unit = VersionConstraintSieve()
        unit.update_configuration(
            {
                "package_name": "tensorboard",
                "version_specifier": ">2.0",
            }
        )
        unit.pre_run()
        assert list(unit.run([package_version])) == [package_version]

    def test_super_pre_run(self, context: Context) -> None:
        """Make sure the pre-run method of the base is called."""
        unit = VersionConstraintSieve()
        unit.update_configuration(
            {
                "package_name": "tensorflow",
                "version_specifier": "==2.3",
            }
        )

        assert unit.unit_run is False
        unit.unit_run = True

        with unit.assigned_context(context):
            unit.pre_run()

        assert (
            unit.unit_run is False
        ), "Unit flag unit_run not reset, is super().pre_run() called in sources when providing pre_run method!?"



#!/usr/bin/env python3
# This script is invoked to pre-process inputs to adviser. As we download the
# request file from Ceph using Argo, relevant input files for adviser need to
# be extracted. They cannot be supplied directly via environment variables as
# the size of the input can be too large (based on the user request).

"""Prepare inputs to the adviser container."""

import json
import os
from typing import Any
from typing import Dict

_DOCUMENT_ID = os.getenv("THOTH_DOCUMENT_ID")
_REQUEST_FILE_DIR = os.getenv("THOTH_ADVISER_REQUEST_FILE_PATH", "/mnt/workdir/")
_ADVISER_WORKDIR = os.getenv("THOTH_ADVISER_WORKDIR", "/opt/app-root/src")
_ADVISER_INPUTS_DIR = os.path.join(_ADVISER_WORKDIR, "input")
_ADVISER_SUBCOMMAND = os.getenv("THOTH_ADVISER_SUBCOMMAND")


def main() -> None:
    """Prepare input files for adviser."""
    os.makedirs(_ADVISER_INPUTS_DIR, exist_ok=False)

    if not _DOCUMENT_ID:
        raise ValueError("Configuration error: no document id supplied")

    request_file_path = os.path.join(_REQUEST_FILE_DIR, f"{_DOCUMENT_ID}.request")
    if not os.path.isfile(request_file_path):
        raise ValueError("Configuration error: no request file to extract inputs from")

    with open(request_file_path, "r") as fp:
        request_file_content = json.load(fp)

    if _ADVISER_SUBCOMMAND == "advise":
        prepare_adviser(request_file_content)
    elif _ADVISER_SUBCOMMAND == "provenance":
        prepare_provenance(request_file_content)
    elif _ADVISER_SUBCOMMAND == "dependency-monkey":
        prepare_dependency_monkey(request_file_content)
    else:
        raise ValueError(f"Unknown adviser subcommand: {_ADVISER_SUBCOMMAND}")


def _prepare_application_stack(request_file_content: Dict[str, Any]) -> None:
    """Prepare Pipfile and Pipfile.lock."""
    pipfile_content = request_file_content["application_stack"]["requirements"]
    with open(os.path.join(_ADVISER_INPUTS_DIR, "Pipfile"), "w") as pf:
        pf.write(pipfile_content)

    pipfile_lock_content = request_file_content["application_stack"]["requirements_lock"] or "null"
    with open(os.path.join(_ADVISER_INPUTS_DIR, "Pipfile.lock"), "w") as pf:
        pf.write(pipfile_lock_content)


def prepare_adviser(request_file_content: Dict[str, Any]) -> None:
    """Prepare inputs for adviser."""
    _prepare_application_stack(request_file_content)

    library_usage = request_file_content["library_usage"]
    with open(os.path.join(_ADVISER_INPUTS_DIR, "library_usage.json"), "w") as f:
        json.dump(library_usage, f, indent=2)

    runtime_environment = request_file_content["runtime_environment"] or None
    with open(os.path.join(_ADVISER_INPUTS_DIR, "runtime_environment.json"), "w") as f:
        json.dump(runtime_environment, f, indent=2)

    constraints_content = request_file_content.get("constraints") or []
    with open(os.path.join(_ADVISER_INPUTS_DIR, "constraints.txt"), "w") as f:
        json.dump(constraints_content, f, indent=2)

    labels = request_file_content.get("labels") or {}
    with open(os.path.join(_ADVISER_INPUTS_DIR, "labels.json"), "w") as f:
        json.dump(labels, f, indent=2)


def prepare_provenance(request_file_content: Dict[str, Any]) -> None:
    """Prepare inputs for provenance-checker."""
    _prepare_application_stack(request_file_content)


def prepare_dependency_monkey(request_file_content: Dict[str, Any]) -> None:
    """Prepare inputs for dependency-monkey."""
    pipeline = request_file_content["pipeline"]
    with open(os.path.join(_ADVISER_INPUTS_DIR, "pipeline.json"), "w") as f:
        json.dump(pipeline, f, indent=2)

    runtime_environment = request_file_content["runtime_environment"] or None
    with open(os.path.join(_ADVISER_INPUTS_DIR, "runtime_environment.json"), "w") as f:
        json.dump(runtime_environment, f, indent=2)

    requirements = request_file_content["requirements"]
    with open(os.path.join(_ADVISER_INPUTS_DIR, "Pipfile"), "w") as f:
        json.dump(requirements, f, indent=2)

    context = request_file_content["context"] or None
    with open(os.path.join(_ADVISER_INPUTS_DIR, "context.json"), "w") as f:
        json.dump(context, f, indent=2)


__name__ == "__main__" and main()  # type: ignore



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A boot that checks for platform used and adjust to the default one if not provided explicitly."""

import logging
from typing import Any
from typing import Dict
from typing import Generator
from typing import TYPE_CHECKING

from thoth.common import get_justification_link as jl

import attr
from voluptuous import Required
from voluptuous import Schema

from ..boot import Boot
from ..exceptions import NotAcceptable

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class PlatformBoot(Boot):
    """A boot that checks for platform used and adjust to the default one if not provided explicitly."""

    CONFIGURATION_DEFAULT = {"default_platform": "linux-x86_64"}
    CONFIGURATION_SCHEMA = Schema(
        {
            Required("default_platform"): str,
        }
    )
    _JUSTIFICATION_LINK = jl("platform")

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Register self, always."""
        if not builder_context.is_included(cls):
            yield {}
            return None

        yield from ()
        return None

    def run(self) -> None:
        """Check for platform configured and adjust to the default one if not provided by user."""
        if self.context.project.runtime_environment.platform is None:
            msg = (
                f"No platform provided in the configuration, setting to "
                f"{self.configuration['default_platform']!r} implicitly"
            )

            _LOGGER.warning("%s - see %s", msg, self._JUSTIFICATION_LINK)
            self.context.project.runtime_environment.platform = self.configuration["default_platform"]
            self.context.stack_info.append({"type": "WARNING", "message": msg, "link": self._JUSTIFICATION_LINK})

        platform = self.context.project.runtime_environment.platform
        if not self.context.graph.python_package_version_depends_on_platform_exists(platform):
            raise NotAcceptable(f"No platform conforming to {platform!r} found in the database")



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A wrap that sorts justifications."""

from typing import TYPE_CHECKING
from typing import Generator
from typing import Dict
from typing import Any

from ..state import State
from ..wrap import Wrap

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext


class SortJustificationsWrap(Wrap):
    """A wrap that sorts justifications."""

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[Any, Any], None, None]:
        """Include this wrap in adviser, once."""
        if not builder_context.is_included(cls):
            yield {}
            return None

        yield from ()
        return None

    def run(self, state: State) -> None:
        """Sort justifications."""
        state.justification.sort(key=lambda i: (i.get("package_name", ""), i["type"], i["message"]))



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test predictor and its core functionality."""

import os
import flexmock
import pytest

from .base import AdviserTestCase


class TestPredictor(AdviserTestCase):
    """Test predictor's core functionality."""

    def test_keep_temperature_history_init(self, predictor_mock_class: type) -> None:
        """Test initialization of temperature history."""
        assert "THOTH_ADVISER_NO_HISTORY" not in os.environ
        predictor = predictor_mock_class()
        assert predictor.keep_history is False, "Temperature history not kept by default"

        flexmock(os)
        os.should_receive("getenv").with_args("THOTH_ADVISER_HISTORY", 0).and_return("0").and_return("1").twice()

        predictor = predictor_mock_class()
        assert predictor.keep_history is False

        predictor = predictor_mock_class()
        assert predictor.keep_history is True

    def test_assign_context(self, predictor_mock_class: type) -> None:
        """Test assigning context to predictor."""
        predictor = predictor_mock_class()
        with pytest.raises(ValueError):
            predictor.context

        context = flexmock()
        with predictor.assigned_context(context):
            assert predictor.context is context



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A step that is assigning scores in a deterministic way."""

from typing import Any
from typing import Optional
from typing import Generator
from typing import Tuple
from typing import List
from typing import Dict
from typing import TYPE_CHECKING
import logging
import random
from pprint import pprint

import attr
from thoth.python import PackageVersion
from voluptuous import Any as SchemaAny
from voluptuous import Optional as SchemaOptional
from voluptuous import Required as SchemaRequired
from voluptuous import Schema

from ...state import State
from ...step import Step


if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext


_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class GenerateScoreStep(Step):
    """A step that is assigning scores in a deterministic way.

    This unit can be used to measure assigning score in a deterministic way across multiple runs without
    a need to store all the score for packages.
    """

    # Assign probability is used to "assign" a score to the package to simulate knowledge
    # coverage for packages resolved - 0.75 means ~75% of packages will have a score.
    CONFIGURATION_SCHEMA: Schema = Schema(
        {
            SchemaOptional("assign_probability"): float,
            SchemaOptional("buffer_size"): int,
            SchemaOptional("package_name"): SchemaAny(str, None),
            SchemaOptional("seed"): int,
            SchemaRequired("multi_package_resolution"): bool,
        }
    )
    CONFIGURATION_DEFAULT: Dict[str, Any] = {
        "assign_probability": 0.75,
        "buffer_size": 1024,
        "multi_package_resolution": False,
        "package_name": None,
        "seed": 42,
    }

    _history = attr.ib(type=Dict[Tuple[str, str, str], float], factory=dict, init=False)
    _buffer = attr.ib(type=List[float], factory=list, init=False)
    _idx = attr.ib(type=int, default=0, init=False)

    def pre_run(self) -> None:
        """Initialize this pipeline units before each run."""
        self._history.clear()
        self._idx = 0

        if not self._buffer:
            state = random.getstate()
            random.seed(self.configuration["seed"])
            self._buffer = [0.0] * self.configuration["buffer_size"]
            for i in range(self.configuration["buffer_size"]):
                self._buffer[i] = (
                    random.uniform(self.SCORE_MIN, self.SCORE_MAX)
                    if random.random() <= self.configuration["assign_probability"]
                    else 0.0
                )
            random.setstate(state)

        super().pre_run()

    def post_run(self) -> None:
        """Print the generated scores on finish to stdout."""
        pprint(self._history)

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Register self, never."""
        yield from ()
        return None

    def run(
        self, _: State, package_version: PackageVersion
    ) -> Optional[Tuple[Optional[float], Optional[List[Dict[str, str]]]]]:
        """Score the given package."""
        package_tuple = package_version.to_tuple()
        score = self._history.get(package_tuple)

        if score is not None:
            return score, None

        idx = self._idx
        self._idx = (self._idx + 1) % self.configuration["buffer_size"]
        self._history[package_tuple] = self._buffer[idx]
        return self._buffer[idx], None
